#include "ap.h"


// Highly parallel three modular multiplications using the FPU
inline void modmul3 (modint &a1, modint &b1, modint &a2, modint &b2, modint &a3, modint &b3)
{
    asm ("pushl %3; pushl %4;
          pushl %5; pushl %6;
          pushl %7; pushl %8;
          fildl 20(%%esp); fildl 16(%%esp);
          fildl 12(%%esp); fxch %%st(2);
          fmulp %%st, %%st(1); fxch %%st(1);
          fildl 8(%%esp); fildl 4(%%esp); fxch %%st(2);
          fmulp %%st, %%st(1); fxch %%st(1);
          fildl (%%esp); addl $12, %%esp;
          fmulp %%st, %%st(1);
          fld %%st(2); fmull imodulus;
          fld %%st(2); fmull imodulus;
          fld %%st(2); fmull imodulus; fxch %%st(2);
          fadds chopper64; fxch %%st(1);
          fadds chopper64; fxch %%st(2);
          fadds chopper64; fxch %%st(1);
          fsubs chopper64; fxch %%st(2);
          fsubs chopper64; fxch %%st(1);
          fsubs chopper64; fxch %%st(2);
          fmull dmodulus; fxch %%st(1);
          fmull dmodulus; fxch %%st(1);
          fsubrp %%st, %%st(5); fxch %%st(1);
          fmull dmodulus; fxch %%st(1);
          fsubrp %%st, %%st(3);
          fsubrp %%st, %%st(1); fxch %%st(2);
          fistpl (%%esp);
          fistpl 4(%%esp);
          fistpl 8(%%esp);
          popl %0;
          popl %1;
          popl %2"
                            : "=rm" (a1), "=rm" (a2), "=rm" (a3)
                            : "0" (a1), "rm" (b1), "1" (a2), "rm" (b2), "2" (a3), "rm" (b3)
                            : "cc");
}

// Partially parallel two modular multiplications using the FPU
inline void modmul2 (modint &a1, modint &b1, modint &a2, modint &b2)
{
    asm ("pushl %2; pushl %3;
          pushl %4; pushl %5;
          fildl 12(%%esp); fildl 8(%%esp);
          fildl 4(%%esp); fxch %%st(2);
          fmulp %%st, %%st(1); fxch %%st(1);
          fildl (%%esp); addl $8, %%esp;
          fmulp %%st, %%st(1);
          fld %%st(1); fmull imodulus;
          fld %%st(1); fmull imodulus; fxch %%st(1);
          fadds chopper64; fxch %%st(1);
          fadds chopper64; fxch %%st(1);
          fsubs chopper64; fxch %%st(1);
          fsubs chopper64; fxch %%st(1);
          fmull dmodulus; fxch %%st(1);
          fmull dmodulus; fxch %%st(1);
          fsubrp %%st, %%st(3);
          fsubrp %%st, %%st(1); fxch %%st(1);
          fistpl (%%esp);
          fistpl 4(%%esp);
          popl %0;
          popl %1"
                            : "=rm" (a1), "=rm" (a2)
                            : "0" (a1), "rm" (b1), "1" (a2), "rm" (b2)
                            : "cc");
}

// The "six-step" fnt, but doesn't transpose or scramble (for convolution only)

void tablesixstepfnttrans (modint data[], modint pr, int isign, size_t nn)
{
    size_t n1, n2, j, k;
    modint w, tmp, tmp2, *p1, *p2;

    if (nn < 2) return;

    for (n1 = 1, n2 = 0; n1 < nn; n1 <<= 1, n2++);
    n1 = n2 >> 1;
    n2 -= n1;

    n1 = 1 << n1;
    n2 = 1 << n2;

    // n2 >= n1

    modint *wtable = new modint[n2];
    size_t *ptable = new size_t[n1];

    if (isign > 0)
        w = pow (pr, modint::modulus - 1 - (modint::modulus - 1) / nn);
    else
        w = pow (pr, (modint::modulus - 1) / nn);

    // treat the input data as a n1 x n2 matrix

    // first transpose the matrix

    transpose (data, n1, n2);

    // then do n2 transforms of length n1

    // init tables

    tmp = pow (w, nn / n1);
    tmp2 = 1;
    for (k = 0; k < n1; k++)
    {
        wtable[k] = tmp2;
        tmp2 *= tmp;
        ptable[k] = permute (k, n1);
    }

    for (k = 0, p1 = data; k < n2; k++, p1 += n1)
        tablefnt (p1, wtable, ptable, n1);

    // transpose the matrix

    transpose (data, n2, n1);

    // then multiply the matrix A_jk by exp(isign * -2 pi i j k / nn)

    tmp = w;
    for (j = 1, p1 = data + n2; j < n1; j++, p1 += n2)
    {
        tmp2 = pow (tmp, j);
        p1[j] *= tmp2;
        tmp2 *= tmp;
        for (k = j + 1, p2 = p1 + n2 + j; k < n1; k++, p2 += n2)
        {
            /*
            p1[k] *= tmp2;
            *p2 *= tmp2;
            tmp2 *= tmp;
            */

            modmul3 (tmp2, tmp, p1[k], tmp2, *p2, tmp2);
        }
        for (; k < n2; k++)
        {
            /*
            p1[k] *= tmp2;
            tmp2 *= tmp;
            */

            modmul2 (tmp2, tmp, p1[k], tmp2);
        }
        tmp *= w;
    }

    // last do n1 transforms of length n2

    // init table

    if (n2 != n1)
    {
        // n2 = 2 * n1
        for (k = n1; k--;)
            wtable[2 * k] = wtable[k];
        tmp = pow (w, nn / n2);
        tmp2 = tmp;
        tmp *= tmp;
        for (k = 1; k < n2; k += 2)
        {
            wtable[k] = tmp2;
            tmp2 *= tmp;
        }
    }

    for (k = 0, p1 = data; k < n1; k++, p1 += n2)
        tablefnt (p1, wtable, 0, n2, 0);

    delete[] ptable;
    delete[] wtable;
}

void itablesixstepfnttrans (modint data[], modint pr, int isign, size_t nn)
{
    size_t n1, n2, j, k;
    modint w, tmp, tmp2, *p1, *p2, inn;

    if (nn < 2) return;

    for (n1 = 1, n2 = 0; n1 < nn; n1 <<= 1, n2++);
    n1 = n2 >> 1;
    n2 -= n1;

    n1 = 1 << n1;
    n2 = 1 << n2;

    // n2 >= n1

    modint *wtable = new modint[n2];
    size_t *ptable = new size_t[n1];

    if (isign > 0)
        w = pow (pr, modint::modulus - 1 - (modint::modulus - 1) / nn);
    else
        w = pow (pr, (modint::modulus - 1) / nn);

    // treat the input data as a n1 x n2 matrix

    // first do n1 transforms of length n2

    // init table

    tmp = pow (w, nn / n2);
    tmp2 = 1;
    for (k = 0; k < n2; k++)
    {
        wtable[k] = tmp2;
        tmp2 *= tmp;
    }

    for (k = 0, p1 = data; k < n1; k++, p1 += n2)
        itablefnt (p1, wtable, 0, n2, 0);

    // then multiply the matrix A_jk by exp(isign * -2 pi i j k / nn) / nn

    tmp = 1;
    inn = modint (1) / nn;
    for (j = 0, p1 = data; j < n1; j++, p1 += n2)
    {
        tmp2 = pow (tmp, j) * inn;
        p1[j] *= tmp2;
        tmp2 *= tmp;
        for (k = j + 1, p2 = p1 + n2 + j; k < n1; k++, p2 += n2)
        {
            /*
            p1[k] *= tmp2;
            *p2 *= tmp2;
            tmp2 *= tmp;
            */

            modmul3 (tmp2, tmp, p1[k], tmp2, *p2, tmp2);
        }
        for (; k < n2; k++)
        {
            /*
            p1[k] *= tmp2;
            tmp2 *= tmp;
            */

            modmul2 (tmp2, tmp, p1[k], tmp2);
        }
        tmp *= w;
    }

    // transpose the matrix

    transpose (data, n1, n2);

    // then do n2 transforms of length n1

    // init table

    if (n2 != n1)
        // n2 = 2 * n1
        for (k = 0; k < n1; k++)
            wtable[k] = wtable[2 * k];

    for (k = 0; k < n1; k++)
        ptable[k] = permute (k, n1);

    for (k = 0, p1 = data; k < n2; k++, p1 += n1)
        itablefnt (p1, wtable, ptable, n1);

    // last transpose the matrix

    transpose (data, n2, n1);

    delete[] ptable;
    delete[] wtable;
}
